# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['mqtt_logger']

package_data = \
{'': ['*']}

install_requires = \
['paho-mqtt>=1.6.1,<2.0.0', 'rich>=12.0.0,<13.0.0']

setup_kwargs = {
    'name': 'mqtt-logger',
    'version': '0.3.2',
    'description': 'Python based MQTT to SQLite3 logger',
    'long_description': '# MQTT to SQLite Logger\n\n[![PyPI version](https://badge.fury.io/py/mqtt-logger.svg)](https://badge.fury.io/py/mqtt-logger)\n[![Python package](https://github.com/Blake-Haydon/mqtt-logger/actions/workflows/python-package.yml/badge.svg)](https://github.com/Blake-Haydon/mqtt-logger/actions/workflows/python-package.yml)\n[![Upload Python Package](https://github.com/Blake-Haydon/mqtt-logger/actions/workflows/python-publish.yml/badge.svg)](https://github.com/Blake-Haydon/mqtt-logger/actions/workflows/python-publish.yml)\n\n## Table of Contents\n- [MQTT to SQLite Logger](#mqtt-to-sqlite-logger)\n  - [Table of Contents](#table-of-contents)\n  - [Description](#description)\n    - [`LOG` Table](#log-table)\n    - [`RUN` Table](#run-table)\n  - [Installation](#installation)\n  - [Example Usage](#example-usage)\n    - [Recording MQTT Messages](#recording-mqtt-messages)\n    - [Playback Recorded MQTT Messages](#playback-recorded-mqtt-messages)\n  - [Unit Tests](#unit-tests)\n\n## Description\n\n`mqtt-logger` allows for asynchronous data logging of MQTT messages to a SQLite database. The SQLite database has two \ntables called `LOG` and `RUN`. The `LOG` table contains the messages that are being logged. The `RUN` table contains \nthe information about the current run of the program.\n\n### `LOG` Table\n\n| ROW NAME  | DESCRIPTION                                            |\n| --------- | ------------------------------------------------------ |\n| ID        | Unique number assigned to each message (ascending int) |\n| RUN_ID    | ID of the current run (ascending int)                  |\n| UNIX_TIME | Time when the message was received                     |\n| TOPIC     | MQTT topic                                             |\n| MESSAGE   | MQTT message received                                  |\n\n### `RUN` Table\n\n| ROW NAME        | DESCRIPTION                                   |\n| --------------- | --------------------------------------------- |\n| ID              | Unique number assigned to run (ascending int) |\n| START_UNIX_TIME | Time when logger was started                  |\n| END_UNIX_TIME   | Time when logger was stopped                  |\n\n\n--- \n\n\n## Installation\n\nIf you are using `mqtt-logger` as a python package, you can install it using pip.\n\n```bash\n# To use as a package\npip install mqtt-logger\n```\n\nIf you are looking to develop `mqtt-logger`, clone and run the following commands (poetry must be installed). \n\n```bash\n# For development work\ngit clone git@github.com:Blake-Haydon/mqtt-logger.git\ngit config --local core.hooksPath .githooks/\npoetry install\n```\n\n---\n\n\n## Example Usage\n\n### Recording MQTT Messages\n\nThis example records messages to the `test/#` topic using a public MQTT broker. It will record for 10 seconds. If you \nare using a private broker, you may need to set the `username` and `password` parameters.\n\n```bash\n# Run example in terminal\npoetry run python examples/10s_recording.py\n```\n\nExample recorder taken from [examples/10s_recording.py](examples/10s_recording.py)\n```python\nimport mqtt_logger\nimport os\nimport time\n\n# Initalise mqtt recorder object\nrec = mqtt_logger.Recorder(\n    sqlite_database_path=os.path.join(os.path.dirname(__file__), "MQTT_log.db"),\n    topics=["test/#"],\n    broker_address="broker.hivemq.com",\n    verbose=True,\n)\n\n# Start the logger, wait 10 seconds and stop the logger\nrec.start()\ntime.sleep(10)\nrec.stop()\n```\n\n### Playback Recorded MQTT Messages\n\nThis example plays back previously recorded MQTT messages from `mqtt_logger.Recorder`. If you are using a private \nbroker, you may need to set the `username` and `password` parameters.\n\n```bash\n# Run example in terminal after running the recorder example\npoetry run python examples/10s_playback.py\n```\n\nExample recorder taken from [examples/10s_playback.py](examples/10s_playback.py)\n```python\nimport mqtt_logger\nimport os\n\n# Initalise playback object\nplayback = mqtt_logger.Playback(\n    sqlite_database_path=os.path.join(os.path.dirname(__file__), "MQTT_log.db"),\n    broker_address="broker.hivemq.com",\n    verbose=True,\n)\n\n# Start playback at 2x speed (twice as fast)\nplayback.play(speed=2)\n```\n\n\n---\n\n\n## Unit Tests\n\n```bash\n# Run tests in poetry virtual environment\npoetry run pytest\n```',
    'author': 'Blake Haydon',
    'author_email': 'blake.a.haydon@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/Blake-Haydon/mqtt-logger',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



import logging
import sqlite3
import asyncio

import paho.mqtt.client as mqtt
from rich.logging import RichHandler

from mqtt_logger.database import (
    create_tables,
    insert_log_entry,
    start_run_entry,
    stop_run_entry,
    tables_exist,
    retrieve_log_entries,
    start_time,
)

# Set logging to output all info by default
logging.basicConfig(
    format="%(message)s",
    level=logging.WARNING,
    handlers=[RichHandler()],
)


class Recorder:
    """Record incoming MQTT messages on selected topics.

    Parameters
    ----------
    sqlite_database_path : str
        Filepath to the *.db file that sqlite uses to save data
    topics : List(str)
        A list containing the topic strings that are to be subscribed to
    broker_address : str
        The IP address that the MQTT broker lives on
    verbose : bool
        Set logging output to INFO level if True
    username : str
        Username for mqtt broker
    password : str
        Password for mqtt broker

    Attributes
    ----------
    topics : list(str)
        List of topic names
    _recording : bool
        Whether the Recorder object is currently recording or not
    _run_id : int
        The id of the current run
    _con : sqlite3.Connection
        Connection object that updates the database file located at sqlite_database_path
    _client : paho.mqtt.client
        MQTT client that connects to the broker and recives the messages
    """

    def __init__(
        self,
        sqlite_database_path: str = "MQTT_log.db",
        topics: list = ["#"],
        broker_address: str = "localhost",
        verbose: bool = False,
        username: str = None,
        password: str = None,
    ):
        # If set to verbose print info messages
        if verbose:
            logging.getLogger().setLevel(logging.INFO)

        # Connect to sqlite database
        # check_same_thread needs to be false as the MQTT callbacks run on a different thread
        self._con = sqlite3.connect(sqlite_database_path, check_same_thread=False)

        # Generate tables if none currently exist
        if tables_exist(self._con):
            logging.warning(
                f"Tables already exist, mqtt_logger will append to {sqlite_database_path}"
            )
        else:
            create_tables(self._con)
            logging.info(f"Created tables in {sqlite_database_path}")

        # The logger object can subscribe to many topics (if none are selected then it will subscribe to all)
        self.topics = topics

        # There is no run id when the object is created, only when the logger is started
        self._run_id = None

        # Do not start logging when object is created (wait for start method)
        self._recording = False

        # Connect to MQTT broker
        self._client = mqtt.Client()
        self._client.on_connect = self._on_connect
        self._client.on_message = self._on_message
        if username is not None and password is not None:
            self._client.username_pw_set(username, password)
        self._client.connect(broker_address)
        self._client.loop_start()  # Threaded execution loop

    def _on_connect(self, client, userdata, flags, rc):
        """Callback function for MQTT broker on connection."""

        if rc == 0:
            logging.info("Connection Successful")
        else:
            raise ConnectionError(
                "Connection was unsuccessful, check that the broker IP is corrrect"
            )

        # Subscribe to all of the topics
        for topic in self.topics:
            self._client.subscribe(topic)
            logging.info(f"Subscribed to: {topic}")

    def _on_message(self, client, userdata, msg):
        """Callback function for MQTT broker on message that logs the incoming MQTT message."""
        if self._recording:
            try:
                insert_log_entry(self._con, msg.topic, msg.payload, self._run_id)
                logging.info(f"{len(msg.payload):>4} bytes <- {msg.topic}")

            except Exception as e:
                logging.error(f"{type(e)}: {e}")

    def start(self):
        """Starts the MQTT logging."""

        if self._recording:
            raise RuntimeError("Already recording, please stop first")

        self._run_id = start_run_entry(self._con)
        self._recording = True

        logging.info(f"Logging started for run {self._run_id}")

    def stop(self):
        """Graceful exit for closing the database connection and stopping the MQTT client."""

        if not self._recording:
            raise RuntimeError("Not recording, please start first")

        stop_run_entry(self._con, self._run_id)
        self._run_id = None
        self._recording = False
        logging.info("Logging stopped")
        self._client.loop_stop()
        logging.info("MQTT broker disconnected")
        self._con.close()
        logging.info("Database connection closed")


class Playback:
    """Playback MQTT messages from log files in realtime or faster.

    Parameters
    ----------
    sqlite_database_path : str
        Filepath to the *.db file that sqlite uses to save data
    broker_address : str
        The IP address of the MQTT broker
    verbose : bool
        Set logging output to INFO level if True
    username : str
        Username for mqtt broker
    password : str
        Password for mqtt broker

    Attributes
    ----------
    _con : sqlite3.Connection
        Connection object that updates the database file located at sqlite_database_path
    _client : paho.mqtt.client
        MQTT client that connects to the broker and recives the messages
    _log_data: list(dict)
        A list of all of the rows in the log file stored in dict format
    """

    def __init__(
        self,
        sqlite_database_path: str = "MQTT_log.db",
        broker_address: str = "localhost",
        verbose: bool = False,
        username: str = None,
        password: str = None,
    ):

        # If set to verbose print info messages
        if verbose:
            logging.getLogger().setLevel(logging.INFO)

        # Connect to sqlite database
        # check_same_thread needs to be false as the MQTT callbacks run on a different thread
        self._con = sqlite3.connect(sqlite_database_path, check_same_thread=False)

        # Retrieve all of the log entries from the database
        self._log_data = retrieve_log_entries(self._con)

        # Connect to MQTT broker
        self._client = mqtt.Client()
        if username is not None and password is not None:
            self._client.username_pw_set(username, password)
        self._client.on_connect = self._on_connect
        self._client.connect(broker_address)
        self._client.loop_start()  # Threaded execution loop

    def _on_connect(self, client, userdata, flags, rc):
        """Callback function for MQTT broker on connection."""

        if rc == 0:
            logging.info("Connection Successful")
        else:
            raise ConnectionError(
                "Connection was unsuccessful, check that the broker IP is corrrect"
            )

    def play(self, speed: float = 1):
        """Play the logged data at a certain speed using an async function.

        Parameters
        ----------
        speed : float
            Speed multiplier to determine how fast to send out the data. A higher value means faster.
        """

        logging.info(f"Playback initiated at {speed}x speed")

        # Run the event loop to issue out all of the MQTT publishes
        asyncio.run(self._publish(speed))

    async def _publish(self, speed: float):
        """Async function that collects all the necessary publish functions and gathers them to then be run by the
        event loop.

        Parameters
        ----------
        speed : float
            Speed multiplier to determine how fast to send out the data. A higher value means faster.
        """
        start_unix_time = start_time(self._con)

        async def _publish_aux(log):
            scaled_sleep = (log["unix_time"] - start_unix_time) / speed
            await asyncio.sleep(scaled_sleep)

            try:
                self._client.publish(log["topic"], log["message"])
                logging.info(f"{len(log['message']):>4} bytes -> {log['topic']}")

            except Exception as e:
                logging.error(f"{type(e)}: {e}")

        # Load all async operation at the start using gather
        publish_queue = [_publish_aux(log) for log in self._log_data]
        await asyncio.gather(*publish_queue, return_exceptions=True)



import sqlite3
import time


LOGGER_TABLE_NAME = "LOG"
RUNS_TABLE_NAME = "RUN"


def create_tables(con: sqlite3.Connection):
    """Initialise the database by creating the necessary tables for logging."""
    cur = con.cursor()

    query = f"""
        CREATE TABLE {LOGGER_TABLE_NAME}
        (ID                         INTEGER             PRIMARY KEY,
        RUN_ID                      INTEGER             NOT NULL,
        UNIX_TIME                   DECIMAL(15,6)       NOT NULL,
        TOPIC                       VARCHAR             NOT NULL,
        MESSAGE                     BLOB                NOT NULL);
    """
    cur.execute(query)

    query = f"""
        CREATE TABLE {RUNS_TABLE_NAME}
        (ID                         INTEGER             PRIMARY KEY,
        START_UNIX_TIME             DECIMAL(15,6)       NOT NULL,
        END_UNIX_TIME               DECIMAL(15,6));
        """
    cur.execute(query)

    # Commit database tables to the database
    con.commit()


def tables_exist(con: sqlite3.Connection) -> bool:
    """Checks if the LOG table exists in the database."""
    cur = con.cursor()

    log_query = f"""
        SELECT name FROM sqlite_master WHERE type='table' AND name='{LOGGER_TABLE_NAME}'
        """

    run_query = f"""
        SELECT name FROM sqlite_master WHERE type='table' AND name='{RUNS_TABLE_NAME}'
        """

    log_exists = cur.execute(log_query).fetchone() is not None
    run_exists = cur.execute(run_query).fetchone() is not None

    if log_exists != run_exists:
        raise RuntimeError(
            "Tables must exist together. Check that the database is either empty or contains both tables."
        )

    return log_exists  # or run_exists could be used as they are equivalent


def start_run_entry(con: sqlite3.Connection) -> int:
    """Inserts a run entry into the database. Returns the run id."""
    cur = con.cursor()

    query = f"""
        INSERT INTO {RUNS_TABLE_NAME}
        (START_UNIX_TIME)
        VALUES ({time.time()})
        """

    cur.execute(query)
    con.commit()
    return cur.lastrowid


def stop_run_entry(con: sqlite3.Connection, run_id: int):
    """Inserts a run entry into the database. Returns the run id."""
    cur = con.cursor()

    query = f"""
        UPDATE {RUNS_TABLE_NAME}
        SET END_UNIX_TIME = {time.time()}
        WHERE ROWID = {run_id}
        """

    cur.execute(query)
    con.commit()


def insert_log_entry(con: sqlite3.Connection, topic: str, message: bytes, run_id: int):
    """Inserts a log entry into the database."""
    cur = con.cursor()

    query = f"""
        INSERT INTO {LOGGER_TABLE_NAME}
        (UNIX_TIME, TOPIC, MESSAGE, RUN_ID)
        VALUES ({time.time()}, ?, ?, ?)
        """

    # NOTE: time.time() will not be the correct time if the system clock is reset (i.e on a raspberry pi)
    cur.execute(query, (topic, message, run_id))
    con.commit()


def retrieve_log_entries(con: sqlite3.Connection):
    """Retrieves all log entries from the database."""
    cur = con.cursor()

    query = f"""
        SELECT UNIX_TIME, TOPIC, MESSAGE FROM {LOGGER_TABLE_NAME}
        """

    # Convert list of tuples into a list of dicts
    return [
        {
            "unix_time": record[0],
            "topic": record[1],
            "message": record[2],
        }
        for record in cur.execute(query).fetchall()
    ]


def start_time(con: sqlite3.Connection) -> float:
    """Retrieves the logging start time"""
    cur = con.cursor()

    # TODO: Implement this with run numbers
    query = f"""
        SELECT MIN(UNIX_TIME) FROM {LOGGER_TABLE_NAME}
        """

    return cur.execute(query).fetchone()[0]
